import abc

from attr import define


@define(kw_only=True, frozen=True)
class Resolver:
    """Resolves credentials."""

    name: str

    @abc.abstractmethod
    def resolve(self, credential: str):
        raise NotImplementedError()



from sqlalchemy import MappingResult, text

from detectpii.catalog import PostgresCatalog
from detectpii.model import Table


class YugabyteCatalog(PostgresCatalog):

    def sample(
        self,
        table: Table,
        percentage: int = 10,
        *args,
        **kwargs,
    ) -> MappingResult:
        sql = text(
            f"""
            SELECT *
            FROM {table.name}
            WHERE RANDOM() <= {percentage / 100}
        """
        )

        with self.engine.connect() as conn:
            return conn.execute(
                statement=sql,
                parameters={
                    "percentage": percentage,
                },
            ).mappings()



from typing import Sequence

from attrs import define
from sqlalchemy import Engine, create_engine, text, Connection, MappingResult

from detectpii.model import Catalog, Table, Column


@define(kw_only=True)
class HiveCatalog(Catalog):
    user: str
    password: str
    database: str
    schema: str
    port: str | int
    host: str

    @property
    def engine(self) -> Engine:
        return create_engine(self.url, connect_args={"auth": "LDAP"})

    def detect_tables(self) -> None:
        with self.engine.connect() as conn:
            for row in self._table_names(conn):
                (table_name,) = row

                self._add_table_to_catalog(
                    table_name=table_name,
                    conn=conn,
                )

    def sample(
        self, table: Table, percentage: int = 10, *args, **kwargs
    ) -> MappingResult:
        sql = text(
            f"""
                SELECT *
                FROM {table.name}
                WHERE RAND() <= {percentage / 100}
            """
        )

        with self.engine.connect() as conn:
            return conn.execute(
                statement=sql,
                parameters={
                    "percentage": percentage,
                },
            ).mappings()

    @property
    def url(self) -> str:
        return f"hive://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"

    def _table_names(self, conn: Connection):
        query = text("""SHOW TABLES""")
        return conn.execute(query)

    def _add_table_to_catalog(self, table_name: str, conn: Connection):
        query = text(f"DESC {table_name}")

        table = Table(
            name=table_name,
            columns=[Column(name=row[0]) for row in conn.execute(query)],
            schema="",
        )

        self.tables.append(table)



from attr import define


@define(kw_only=True, frozen=True)
class PiiType:
    """A type of PII"""

    name: str
    type: str


@define(kw_only=True)
class Phone(PiiType):
    name: str = "Phone"
    type: str = "phone"


@define(kw_only=True)
class Email(PiiType):
    name: str = "Email"
    type: str = "email"


@define(kw_only=True)
class CreditCard(PiiType):
    name: str = "Credit Card"
    type: str = "credit_card"


@define(kw_only=True)
class Address(PiiType):
    name: str = "Address"
    type: str = "address"


@define(kw_only=True)
class Person(PiiType):
    name: str = "Person"
    type: str = "person"


@define(kw_only=True)
class BirthDate(PiiType):
    name: str = "Birth Date"
    type: str = "birth_date"


@define(kw_only=True)
class Gender(PiiType):
    name: str = "Gender"
    type: str = "gender"


@define(kw_only=True)
class Nationality(PiiType):
    name: str = "Nationality"
    type: str = "nationality"


@define(kw_only=True)
class SSN(PiiType):
    name: str = "SSN"
    type: str = "ssn"


@define(kw_only=True)
class ZipCode(PiiType):
    name: str = "Zip Code"
    type: str = "zip_code"


@define(kw_only=True)
class PoBox(PiiType):
    name: str = "PO Box"
    type: str = "po_box"


@define(kw_only=True)
class UserName(PiiType):
    name: str = "User Name"
    type: str = "user_name"


@define(kw_only=True)
class Password(PiiType):
    name: str = "Password"
    type: str = "password"


@define(kw_only=True)
class IPAddress(PiiType):
    name: str = "IP Address"
    type: str = "ip_address"



from typing import Optional

from attr import define

from detectpii.common_regex import CommonRegex
from detectpii.detector.column_name_regex import Detector
from detectpii.model import Column
from detectpii.pii_type import (
    PiiType,
    Phone,
    Email,
    IPAddress,
    CreditCard,
    Address,
    ZipCode,
    PoBox,
    SSN,
)


@define(kw_only=True)
class ColumnValueRegexDetector(Detector):
    """Detect PII columns by matching them against known patterns.

    This class has been borrowed from piicatcher and modified for use with this library.
    """

    def detect(
        self,
        column: Column,
        **kwargs,
    ) -> Optional[PiiType]:
        sample = kwargs["sample"]

        if not sample:
            return

        regex = CommonRegex(str(sample))

        if regex.phones or regex.phones_with_exts:  # noqa
            return Phone()

        if regex.emails:  # noqa
            return Email()

        if regex.ips or regex.ipv6s:  # noqa
            return IPAddress()

        if regex.credit_cards:  # noqa
            return CreditCard()

        if regex.street_addresses:  # noqa
            return Address()

        if regex.zip_codes:  # noqa
            return ZipCode()

        if regex.po_boxes:  # noqa
            return PoBox()

        if regex.ssn_number:  # noqa
            return SSN()

        return None



from cattrs import Converter
from cattrs.strategies import configure_tagged_union

from detectpii.catalog import CatalogT
from detectpii.pipeline import PiiDetectionPipeline
from detectpii.resolver import ResolverT
from detectpii.scanner import ScannerT


def pipeline_to_dict(pipeline: PiiDetectionPipeline) -> dict:
    """Convert a pipeline to a dictionary."""

    c = Converter()
    configure_tagged_union(ScannerT, c)
    configure_tagged_union(ResolverT, c)
    configure_tagged_union(CatalogT, c)

    return c.unstructure(pipeline)


def dict_to_pipeline(dictionary: dict) -> PiiDetectionPipeline:
    """Convert a dictionary to a pipeline."""

    c = Converter()
    configure_tagged_union(ScannerT, c)
    configure_tagged_union(ResolverT, c)
    configure_tagged_union(CatalogT, c)

    return c.structure(dictionary, PiiDetectionPipeline)



from attr import define

from detectpii.detector import ColumnNameRegexDetector
from detectpii.detector.column_name_regex import Detector
from detectpii.model import PiiColumn, Scanner, Catalog


@define(kw_only=True)
class MetadataScanner(Scanner):
    """Scan the table schema for PII columns."""

    column_name_regex_detector: Detector = ColumnNameRegexDetector()

    def scan(
        self,
        catalog: Catalog,
        **kwargs,
    ) -> list[PiiColumn]:
        pii_columns = []

        for table in catalog.tables:
            for column in table.columns:
                pii_type = self.column_name_regex_detector.detect(column)

                if pii_type:
                    pii_column = PiiColumn(
                        table=table.name,
                        column=column.name,
                        pii_type=pii_type,
                    )

                    pii_columns.append(pii_column)

        return pii_columns



import abc
from typing import Sequence

from attr import define, Factory
from sqlalchemy import Engine

from detectpii.pii_type import PiiType
from detectpii.resolver import ResolverT, PlaintextResolver


@define(kw_only=True)
class Column:
    """A column in a table."""

    name: str


@define(kw_only=True)
class Table:
    """A table in the database or data warehouse."""

    name: str
    schema: str
    columns: list[Column]


@define(kw_only=True)
class Catalog:
    """A collection of tables."""

    tables: list[Table] = Factory(list)
    resolver: ResolverT = PlaintextResolver()

    @abc.abstractmethod
    def engine(self) -> Engine:
        raise NotImplementedError()

    @abc.abstractmethod
    def detect_tables(self) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def sample(self, table: Table, *args, **kwargs) -> Sequence[dict]:
        raise NotImplementedError()


@define(kw_only=True)
class Scanner:
    """A PII scanner."""

    @abc.abstractmethod
    def scan(self, catalog: Catalog, **kwargs):
        raise NotImplementedError()


@define(kw_only=True, frozen=True)
class PiiColumn:
    """A column that may potentially store PII data."""

    table: str
    column: str
    pii_type: PiiType



from .trino_catalog import TrinoCatalog
from .postgres_catalog import PostgresCatalog
from .yugabyte_catalog import YugabyteCatalog
from .snowflake_catalog import SnowflakeCatalog
from .hive_catalog import HiveCatalog

CatalogT = (
    PostgresCatalog | TrinoCatalog | YugabyteCatalog | SnowflakeCatalog | HiveCatalog
)



import re

from attr import define, Factory
from detectpii.catalog import CatalogT
from detectpii.model import PiiColumn
from detectpii.scanner import ScannerT


@define(kw_only=True)
class PiiDetectionPipeline:
    catalog: CatalogT
    scanners: list[ScannerT] = Factory(list)
    ignore_column_regex: str | None = None

    def scan(self) -> list[PiiColumn]:
        self.catalog.detect_tables()
        pii_columns = []

        for scanner in self.scanners:
            pii_columns.extend(scanner.scan(catalog=self.catalog))

        if self.ignore_column_regex:
            return [
                pii_column
                for pii_column in pii_columns
                if not re.match(self.ignore_column_regex, pii_column.column)
            ]

        return list(set(pii_columns))
